import {
  assertRequestData
} from "./chunk-Z4RMEK24.js";

// src/createClient.ts
import WebSocket from "isomorphic-ws";
import { pack, unpack } from "msgpackr";

// src/utils/createIntegroPromise.ts
var _args = Symbol("args");
var _setPromise = Symbol("setPromise");
var createIntegroPromise = (executor, ...args) => {
  let promise;
  return {
    [Symbol.toStringTag]: "IntegroPromise",
    [_args]: args,
    [_setPromise]: (nativePromise) => {
      promise ??= nativePromise;
    },
    catch: (onrejected) => {
      promise ??= Promise.resolve(executor(...args));
      return promise.catch(onrejected);
    },
    finally: (onfinally) => {
      promise ??= Promise.resolve(executor(...args));
      return promise.finally(onfinally);
    },
    then: (onfulfilled, onrejected) => {
      promise ??= Promise.resolve(executor(...args));
      return promise.then(onfulfilled, onrejected);
    }
  };
};
var getIntegroPromiseArgs = (promise) => promise[_args];
var setIntegroPromiseValue = (promise, nativePromise) => {
  promise[_setPromise]?.(nativePromise);
};
var isIntegroPromise = (promise) => !!promise && typeof promise === "object" && promise.toString() === "[object IntegroPromise]" && Array.isArray(promise[_args]);
var isIntegroPromises = (promises) => promises.every(isIntegroPromise);

// src/utils/noop.ts
var noop = () => {
};

// src/utils/createProxy.ts
var createProxy = (apply, path = []) => new Proxy(noop, {
  get: (_target, key) => key === Symbol.toStringTag ? path.join(".") : createProxy(apply, [...path, key.toString()]),
  apply: (_target, _thisArg, args) => apply(path, args)
});

// src/utils/isArrayEqual.ts
var isArrayEqual = (a, b) => a.length === b.length && a.every((value, index) => value === b[index]);

// src/createClient.ts
var post = async (options) => {
  const { url, config } = Array.isArray(options) ? options[0] : options;
  const data = Array.isArray(options) ? options.map((o) => o.data) : options.data;
  const init = typeof config.requestInit === "function" ? config.requestInit() : config.requestInit;
  const auth = typeof config.auth === "function" ? config.auth() : config.auth;
  const res = await fetch(url, {
    method: "POST",
    ...init,
    headers: {
      "Content-Type": "application/msgpack",
      ...auth ? { "Authorization": auth } : void 0,
      ...init?.headers
    },
    body: pack(data)
  });
  const arrayBuffer = await res.arrayBuffer();
  const unpacked = unpack(new Uint8Array(arrayBuffer));
  if (!res.ok) {
    if (typeof unpacked === "object" && unpacked !== null && Object.getPrototypeOf(unpacked) === Object.prototype && "message" in unpacked && typeof unpacked.message === "string" && unpacked.message.length) {
      throw new Error(unpacked.message);
    }
    throw new Error("The server responded in error.", {
      cause: unpacked
    });
  }
  return unpacked;
};
var subscribe = ({
  url,
  config,
  data: { path, args },
  websocket
}) => {
  const handler = args[args.length - 1];
  const sendArgs = args.slice(0, args.length - 1);
  if (typeof handler !== "function")
    throw new Error("Last parameter must be a callback function.");
  const listener = (data) => {
    const unpacked = unpack(new Uint8Array(data));
    if (unpacked.type === "error") {
      handler(void 0, unpacked.message);
      websocket.current?.off("message", listener);
    }
    ;
    if (unpacked.type !== "event")
      return;
    if (!isArrayEqual(unpacked.path, path))
      return;
    handler(unpacked.message);
  };
  websocket.current ??= new WebSocket(url);
  websocket.current.on("error", console.error);
  websocket.current.on("message", listener);
  websocket.isOpen ? websocket.current?.send(pack({ type: "subscribe", auth: config.auth, path, args: sendArgs })) : websocket.current.on("open", () => {
    websocket.isOpen = !!websocket.current;
    websocket.current?.send(pack({ type: "subscribe", auth: config.auth, path, args: sendArgs }));
  });
  return () => {
    websocket.current?.off("message", listener);
  };
};
var createClient = (url = "/", config = {}) => {
  const websocket = { current: void 0, isOpen: false };
  return createProxy(
    (path, args) => path.length >= 2 && path[path.length - 1] === (config.subscribeKey ?? "subscribe") ? subscribe({
      url,
      config,
      data: {
        path,
        args
      },
      websocket
    }) : createIntegroPromise(post, { url, config, data: { type: "request", path, args } })
  );
};

// src/batch.ts
var assertPostOptions = (value) => {
  if (!Array.isArray(value) || !value[0] || typeof value[0] !== "object" || typeof value[0].url !== "string" || !value[0].config || typeof value[0].config !== "object") {
    throw new Error("Only IntegroPromises can be batched");
  }
  try {
    assertRequestData(value[0].data);
  } catch {
    throw new Error("Only IntegroPromises can be batched");
  }
  return value.slice(0, 1);
};
var batch = (type) => (promises) => {
  if (!isIntegroPromises(promises))
    throw new Error("Only IntegroPromises can be batched");
  const allPostOptions = promises.map(
    (promise) => assertPostOptions(getIntegroPromiseArgs(promise))[0]
  );
  const executor = (options) => post(options).then((results) => {
    results.forEach((result, index) => {
      if (type === "all" || type === "allSequential") {
        setIntegroPromiseValue(promises[index], Promise.resolve(result));
      } else {
        const settledResult = result;
        setIntegroPromiseValue(promises[index], settledResult.status === "fulfilled" ? Promise.resolve(settledResult.value) : Promise.reject(settledResult.reason));
      }
    });
    return results;
  });
  const batchPromise = promises.length === 0 ? createIntegroPromise(() => []) : createIntegroPromise(executor, {
    url: allPostOptions[0].url,
    config: allPostOptions[0].config,
    data: {
      type,
      data: allPostOptions.map(({ data }) => data)
    }
  });
  Object.assign(batchPromise, {
    *[Symbol.iterator]() {
      for (const promise of promises) {
        yield promise;
      }
      yield () => batchPromise;
    }
  });
  return batchPromise;
};
var all = batch("all");
var allSequential = batch("allSequential");
var allSettled = batch("allSettled");
var allSettledSequential = batch("allSettledSequential");
export {
  all,
  allSequential,
  allSettled,
  allSettledSequential,
  createClient
};
